// =============================================================================================================================//
// Модуль для работы с MQTT-серверами, "обертка" для PubSubClient                                                               //
// =============================================================================================================================//
// Автор: Александр Разживин, kotyara12@yandex.ru, http://ravsoft2004.narod.ru                                                  //
// =============================================================================================================================//

#include "Arduino.h"
#include "rMQTT.h"
#include "QueueArray.h"
#include "rEEPROM.h"
#include "rNTP.h"
#include <string.h>
#include <TimeLib.h>
#include <ESP8266WiFi.h>
#include <PubSubClient.h>

const long      waitPublMqtt1       = 10;                  // Задержка при публикации только в MQTT без ошибок
const long      waitPublMqtt2       = 2500;                // Задержка при публикации только в MQTT после ошибки
const long      waitPublMqtt3       = 5000;                // Задержка при публикации только в MQTT после второй ошибки
const uint16_t  intvGatewayAct      = 180;                 // Интервал контроля активности шлюза

const bool      retSensors          = true;                // Retained для сенсоров
const bool      retDeviceResp       = true;                // Retained для ответов статуса устройства
const bool      retDeviceSync       = false;               // Retained для запросов статуса устройства
const bool      retSettings         = true;                // Retained для параметров устройства
const byte      qosSensors          = 1;                   // QoS для сенсоров "внутреннего" сервера
const byte      qosDeviceResp       = 2;                   // QoS для ответов статуса устройства
const byte      qosDeviceSync       = 2;                   // QoS для запросов статуса устройства
const byte      qosSettings         = 2;                   // QoS для параметров устройства

const int       queueSizeThingSpeak = 32;                  // Размер очереди на отправку сообщений в ThingSpeak

// Параметры ThingSpeak
const long      waitPublThingSpeak  = 20000;               // Задержка при публикации на сервере ThingSpeak

const char*     ts_ApiServer        = "api.thingspeak.com";
const char*     ts_ApiPost          = "GET /update?%s HTTP/1.1\r\n";
const char*     ts_ApiHost          = "Host: %s\r\n";
const char*     ts_ApiKey           = "api_key=";
const char*     ts_ApiField         = "&field";

// Адреса параметров в EEPROM
const int16_t   addrSyncRequests    = 0;
const int16_t   addrIntvPublishSens = 1;
const int16_t   addrPublThingSpeak  = 3;
const int16_t   addrGatewayEnabled  = 4;

class pubThingSpeakQueue {
  public:
    pubThingSpeakQueue(String _request) {
      request = _request;
    };
  String request;
};

class pubMqttQueue {
  public:
    pubMqttQueue(String _topic, String _payload, bool _retain, byte _qos) {
      topic = _topic;
      payload = _payload;
      retain = _retain;
      qos = _qos;
    };
  String topic;
  String payload;
  bool retain;
  byte qos;
};

WiFiClient _mqttWiFi;
PubSubClient _mqttClient(_mqttWiFi);
QueueArray<struct pubThingSpeakQueue*> _queueThingSpeak(queueSizeThingSpeak);

rLed*           _led;

String          _mqttServer;
int             _mqttPort;
String          _mqttUser;
String          _mqttPass;
bool            _mqttStartSlash;
String          _mqttSite;
String          _mqttDeviceId;

bool            _syncRequests = true;
bool            _publThingSpeak = true;
uint16_t        _intvPublishSensors = 300;
unsigned long   _lastPublishThingSpeak = 0;
unsigned long   _lastPublishSensors = 0;

t_cbPubSensors  _cbPubSensors;
t_cbSubSettings _cbSubSettings;

void _mqttCallback(const MQTT::Publish& pub);
void _publishDeviceState();

void mqttBegin(rLed &led, const char* sServer, const uint16_t iPort, const char* sUser, const char* sPass, const char* sSite, const char* sDeviceId, const bool bStartSlash)
{
  _led = &led;

  _mqttServer = sServer;
  _mqttPort = iPort;
  _mqttUser = sUser;
  _mqttPass = sPass;
  _mqttSite = sSite;
  _mqttDeviceId = sDeviceId;
  _mqttStartSlash = bStartSlash;
  _mqttClient.set_server(_mqttServer, _mqttPort);
  _mqttClient.set_callback(_mqttCallback);
}

// =============================================================================================================================
// == Генерация топиков ========================================================================================================
// =============================================================================================================================
String mqttGetSite()
{
  if (_mqttStartSlash) {
    return String(mqttDiv) + _mqttSite;  
  }
  else {
    return _mqttSite;
  };
}

String mqttGetPrefix() 
{
  return mqttGetSite() + mqttDiv + _mqttDeviceId;
}

String mqttGetTopic(const String topic1) 
{
  return mqttGetPrefix() + mqttDiv + topic1;
}

String mqttGetTopic(const String topic1, const String topic2) 
{
  return mqttGetPrefix() + mqttDiv + topic1 + mqttDiv + topic2;
}

String mqttGetTopic(const String topic1, const String topic2, const String topic3) 
{
  return mqttGetPrefix() + mqttDiv + topic1 + mqttDiv + topic2 + mqttDiv + topic3;
}

String mqttGetTopic(const String topic1, const String topic2, const String topic3, const String topic4) 
{
  return mqttGetPrefix() + mqttDiv + topic1 + mqttDiv + topic2 + mqttDiv + topic3 + mqttDiv + topic4;
}

String mqttGetGlobalTopic(const String topic1, const String topic2) 
{
  return mqttGetSite() + mqttDiv + topic1 + mqttDiv + topic2;
}

String mqttGetGlobalTopic(const String topic1, const String topic2, const String topic3) 
{
  return mqttGetSite() + mqttDiv + topic1 + mqttDiv + topic2 + mqttDiv + topic3;
}

String mqttGetGlobalTopic(const String topic1, const String topic2, const String topic3, const String topic4) 
{
  return mqttGetSite() + mqttDiv + topic1 + mqttDiv + topic2 + mqttDiv + topic3 + mqttDiv + topic4;
}

// =============================================================================================================================
// == Подключение к серверу ====================================================================================================
// =============================================================================================================================
bool mqttConnect() 
{
  bool result = false;
  
  if (_mqttClient.connected()) {
    result = true;
  }
  else {
    _led->ledOn();
    
    // Восстанавливаем подключение к MQTT
    while ((WiFi.status() == WL_CONNECTED) && (!_mqttClient.connected())) {
      // Пробуем подключиться к MQTT брокеру
      Serial.println();
      Serial.println(F("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"));
      Serial.printf("Connecting to MQTT server '%s:%s@%s:%d'...\n", _mqttUser.c_str(), _mqttPass.c_str(), _mqttServer.c_str(), _mqttPort);

      MQTT::Connect mqttCon = MQTT::Connect(_mqttDeviceId);
      mqttCon.set_auth(_mqttUser, _mqttPass);
      mqttCon.set_will(mqttGetTopic(mqttSyncStatus).c_str(), mqttSyncWill);
      
      if (_mqttClient.connect(mqttCon)) {
        Serial.println(F("Connection completed"));
        Serial.println(F("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"));
        Serial.println();

        // Публикуем текущий статус на сервере
        _publishDeviceState();
        Serial.println();

        // Включаем обработку подписок на сервере 1
        mqttSubscribe(mqttGetTopic(mqttSyncStatus).c_str());
        mqttSubscribe(mqttGetTopic(mqttSettings, mqttPlus).c_str());
        mqttSubscribe(mqttGetTopic(mqttSettings, mqttPlus, mqttPlus).c_str());
        Serial.println();
        
        result = true;
      }
      else {
        Serial.println(F("Connection failed, try again in 1 second..."));
        Serial.println(F("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"));
        Serial.println();

        _led->ledToggle();
        delay(1000);
      }
    };

    _led->ledOff();
  };

  return result;
}

bool mqttForceReconnect()
{
  _mqttClient.disconnect();
  return mqttConnect();
}

// =============================================================================================================================
// == ThingSpeak: публикация значений ========================================================================================== 
// =============================================================================================================================
bool _tsPublish(const byte num, const String request)
{
  bool result = false;
 
  // Для отправки запроса на ThingSpeak используем локальный экземпляр WiFiClient
  WiFiClient tsClient;

  _led->ledOn();

  Serial.printf("ThingSpeak %03d:: send data \"%s\"\n", num, request.c_str());

  // Отправляем HTTP-запрос
  if (tsClient.connect(ts_ApiServer, 80)) {
    tsClient.printf(ts_ApiPost, request.c_str());
    tsClient.printf(ts_ApiHost, ts_ApiServer);
    tsClient.println(F("User-Agent: ESP8266 (nothans)/1.0"));
    tsClient.println(F("Connection: close"));
    tsClient.println();

    // Читаем результат (первую строку отправленного заголовка)
    Serial.print(F("ThingSpeak :: response code: "));
    Serial.println(tsClient.readStringUntil('\n'));

    result = true;

    tsClient.stop();
  }
  else {
    Serial.print(F("#ERROR# Failed connection to "));
    Serial.println(ts_ApiServer);
  };

  _led->ledOff();

  return result;
}

void _tsProcessQueue()
{
  if (_publThingSpeak && (WiFi.status() == WL_CONNECTED) && (!_queueThingSpeak.isEmpty()) && ((millis() - _lastPublishThingSpeak) > waitPublThingSpeak)) 
  {
    _lastPublishThingSpeak = millis();

    pubThingSpeakQueue *nextPub = _queueThingSpeak.peek();
    if (_tsPublish(_queueThingSpeak.count(), nextPub->request)) {
      nextPub = _queueThingSpeak.pop();
      delete nextPub;
    };
  };
}

void tsPublish(const String request)
{
  if (_publThingSpeak) 
  {
    // Добавляем в очередь на отправку
    if (_queueThingSpeak.push(new pubThingSpeakQueue(request))) {
      Serial.printf("ThingSpeak :: message \"%s\" :: added to queue, total %d messages\n", request.c_str(), _queueThingSpeak.count());
    }
    else {
      Serial.printf("ThingSpeak :: failed post message '%s' - queue is full!\n", request.c_str());
    };
  };
}

// =============================================================================================================================
// == Публикация сообщений в MQTT ==============================================================================================
// =============================================================================================================================
bool _publish(const String topic, const String payload, const boolean retain, const byte qos) 
{
  bool result = false;

  _led->ledOn();

  MQTT::Publish mqttPub = MQTT::Publish(topic, payload);
  mqttPub.set_retain(retain);
  mqttPub.set_qos(qos);

  result = _mqttClient.publish(mqttPub);

  if (result) {
    Serial.printf("MQTT :: publish '%s': \"%s\" [%d,%d] :: ok\n", topic.c_str(), payload.c_str(), retain, qos);
  }
  else {
    Serial.printf("MQTT :: publish '%s': \"%s\" [%d,%d] :: publish failed!\n", topic.c_str(), payload.c_str(), retain, qos);
  };

  _led->ledOff();
  
  return result;
}

bool mqttPublish(const String topic, const String payload, const boolean retain, const byte qos)
{
  bool result = false;
  
  if (WiFi.status() == WL_CONNECTED) {
    // Если соединения нет, то подключаемся
    if (!_mqttClient.connected()) mqttConnect(); 

    // Отправляем сообщение 
    result = _publish(topic, payload, retain, qos);

    // Если попытка не удалась, пробуем подождать и отправить еще раз
    if (!result) {
      delay(waitPublMqtt2);
      result = _publish(topic, payload, retain, qos);
    }

    // Если попытка не удалась, пробуем переподключиться и еще раз отправить
    if (!result) {
      delay(waitPublMqtt3);
      result = mqttForceReconnect() && _publish(topic, payload, retain, qos);
    };
  };

  if (result) delay(waitPublMqtt1);
  return result;
}

// =============================================================================================================================
// == Подписка на команды ======================================================================================================
// =============================================================================================================================

void mqttSubscribe(const String topic) 
{
  Serial.printf("MQTT :: subscribe on topic '%s'\n", topic.c_str());

  if (!_mqttClient.subscribe(topic)) {
    Serial.println(F("ERROR :: subscribe failed!"));  
  };
}

void mqttUnsubscribe(const String topic) 
{
  Serial.printf("MQTT :: unsubscribe from topic '%s'\n", topic.c_str());

  if (!_mqttClient.unsubscribe(topic)) {
    Serial.println(F("ERROR :: unsubscribe failed!"));  
  };
}

// =============================================================================================================================
// == Статус устройства ========================================================================================================
// =============================================================================================================================
void _publishDeviceState()
{
  mqttPublish(mqttGetTopic(mqttSyncStatus), mqttSyncOk, retDeviceResp, qosDeviceResp);  
}

void _callbackDeviceSync(const String topicName, const String topicValue)
{
  Serial.printf("SYNC :: sync request '%s' \"%s\"\n", topicName.c_str(), topicValue.c_str());
  
  _publishDeviceState();
}

// =============================================================================================================================
// == Публикация данных с сенсоров =============================================================================================
// =============================================================================================================================
void mqttPublishSensor(const String snsTopic, const String itemTopic, const String itemValue)
{
  mqttPublish(mqttGetTopic(snsTopic, itemTopic), itemValue, retSensors, qosSensors);
}

void mqttPublishSensor(const String snsTopic, const String itemTopic, const char* tsKeyAPI, const byte tsFieldNo, const String itemValue)
{
  // Публикуем полученные данные на MQTT брокере
  mqttPublishSensor(snsTopic, itemTopic, itemValue);

  // Публикуем полученные данные на ThingSpeak
  if (_publThingSpeak && (tsKeyAPI != "")) {
    // Формируем GET запрос
    String payload = String(ts_ApiKey) + tsKeyAPI;
    payload = payload + ts_ApiField + String(tsFieldNo) + "=" + itemValue;
    // Отправляем запрос в очередь отправки на сервер
    tsPublish(payload.c_str());
  };

  Serial.println();
}


void mqttPublishSensorData(const String snsTopic, const char* tsKeyAPI, const sensorData &snsData)
{
  mqttPublishSensor(snsTopic, mqttSensorName, snsData.sensorName);
  mqttPublishSensor(snsTopic, mqttSensorType, snsData.sensorType);
  mqttPublishSensor(snsTopic, mqttSensorState, snsData.sensorState);
  mqttPublishSensor(snsTopic, mqttSensorTimeI, String(snsData.sensorTime));
  mqttPublishSensor(snsTopic, mqttSensorTimeS, timeTimeShortString(snsData.sensorTime));
  mqttPublishSensor(snsTopic, mqttSensorTimeL, timeDateTimeString(snsData.sensorTime));
    
  // Публикуем полученные данные на MQTT брокере
  for (byte i = 0; i < snsData.countValues; i++) {
    mqttPublishSensor(snsTopic, snsData.arrayValues[i].topic, snsData.arrayValues[i].value);
  };
  
  // Публикуем полученные данные на ThingSpeak
  if (_publThingSpeak && (tsKeyAPI != "")) {
    // Формируем GET запрос
    String payload = String(ts_ApiKey) + tsKeyAPI;
    for (byte i = 0; i < snsData.countValues; i++) {
      payload = payload + ts_ApiField + String(i + 1) + "=" + snsData.arrayValues[i].value;
    }
    // Отправляем запрос в очередь отправки на сервер
    tsPublish(payload.c_str());
  };

  Serial.println();
}

void mqttPublishDeviceInfo()
{
  mqttPublishSensor(mqttDeviceInfo, mqttChipId, String(ESP.getChipId(), HEX) + " / " + String(ESP.getChipId()));
  mqttPublishSensor(mqttDeviceInfo, mqttCpuFreqMHz, String(ESP.getCpuFreqMHz()));
  mqttPublishSensor(mqttDeviceInfo, mqttFlashChipId, String(ESP.getFlashChipId(), HEX) + " / " + String(ESP.getFlashChipId()));
  mqttPublishSensor(mqttDeviceInfo, mqttFlashChipRealSize, String(ESP.getFlashChipRealSize()));
  mqttPublishSensor(mqttDeviceInfo, mqttFlashChipSpeed, String(ESP.getFlashChipSpeed()));
  mqttPublishSensor(mqttDeviceInfo, mqttIp, WiFi.localIP().toString());
  mqttPublishSensor(mqttDeviceInfo, mqttRSSI, String(WiFi.RSSI()));
  mqttPublishSensor(mqttDeviceInfo, mqttTime, String(now()));
  mqttPublishSensor(mqttDeviceInfo, mqttTimeS, timeTimeShortString());
  mqttPublishSensor(mqttDeviceInfo, mqttTimeL, timeDateTimeString());
  mqttPublishSensor(mqttDeviceInfo, mqttHeapFree, String(ESP.getFreeHeap()));
  mqttPublishSensor(mqttDeviceInfo, mqttHeapFrag, String(ESP.getHeapFragmentation()));
  // mqttPublishSensor(mqttDeviceInfo, mqttPowerVoltage, String(ESP.getVCC()));
}

void mqttSetCallbackSensors(t_cbPubSensors cb)
{
  _cbPubSensors = cb;
}

void _doCallbackSensors()
{
  if (_cbPubSensors != nullptr) {
    _cbPubSensors();
  }
}

void mqttPublishSensors()
{
  // Собираем данные с серверов
  mqttPublishDeviceInfo();
  _doCallbackSensors();

  _lastPublishSensors = millis();
}

// =============================================================================================================================
// == Параметры ================================================================================================================
// =============================================================================================================================
void mqttReadSettings()
{
  Serial.print(F("syncRequests = "));
  _syncRequests = EEPROM_bool_read(addrSyncRequests, _syncRequests);
  Serial.println(_syncRequests);

  Serial.print(F("intvPublishSensors = "));
  _intvPublishSensors = EEPROM_uint16_read(addrIntvPublishSens, _intvPublishSensors);
  Serial.println(_intvPublishSensors);

  Serial.print(F("publThingSpeak = "));
  _publThingSpeak = EEPROM_bool_read(addrPublThingSpeak, _publThingSpeak);
  Serial.println(_publThingSpeak);
}

void mqttSetCallbackSub(t_cbSubSettings cb)
{
  _cbSubSettings = cb;
}

void _doCallbackSub(const String topicName, const String topicValue)
{
  if (_cbSubSettings != nullptr) {
    _cbSubSettings(topicName, topicValue);
  }
}

void _callbackSettings(const String topicName, const String topicValue)
{
  bool processed = false;

  Serial.printf("MQTT :: message arrived '%s' \"%s\"\n", topicName.c_str(), topicValue.c_str());

  // Режим синхронизации
  if (topicName.equals(mqttGetTopic(mqttSettings, mqttSyncMode))) {
    processed = true;
    _syncRequests = (topicValue.equalsIgnoreCase(mqttTrue) || topicValue.equalsIgnoreCase(mqttOn));
    EEPROM_bool_write(addrSyncRequests, _syncRequests);
    Serial.print(F("SETTINGS :: set value 'syncRequests' = "));
    Serial.println(_syncRequests);
  };
  
  // Интервал публикации данных
  if (!processed && topicName.equals(mqttGetTopic(mqttSettings, mqttPubInterval))) {
    processed = true;
    _intvPublishSensors = topicValue.toInt();
    EEPROM_uint16_write(addrIntvPublishSens, _intvPublishSensors);
    Serial.print(F("SETTINGS :: set value 'intvPublishSensors' = "));
    Serial.println(_intvPublishSensors);
  };

  // Режим ThingSpeak
  if (!processed && topicName.equals(mqttGetTopic(mqttSettings, mqttThingSpeak))) {
    processed = true;
    _publThingSpeak = (topicValue.equalsIgnoreCase(mqttTrue) || topicValue.equalsIgnoreCase(mqttOn));
    EEPROM_bool_write(addrPublThingSpeak, _publThingSpeak);
    Serial.print(F("SETTINGS :: set value 'publThingSpeak' = "));
    Serial.println(_publThingSpeak);
  };

  if (!processed && topicName.equals(mqttGetTopic(mqttSettings, mqttCommands))) {
    processed = true;
    if (topicValue.equalsIgnoreCase(mqttCmdReset)) {
      Serial.println(F("Reset ESP!!!"));
      ESP.reset();
      return;
    } 
    else {
      Serial.print(F("Команда '"));
      Serial.print(topicValue);
      Serial.println(F("' не обработана!"));
    };
  };

  if (!processed) {
    _doCallbackSub(topicName, topicValue);
  };
}

// =============================================================================================================================
// == Встроенные функции обратного вызова для PubSubClient =====================================================================
// =============================================================================================================================
void _mqttCallback(const MQTT::Publish& pub)
{
  _led->ledOn();

  String topic = pub.topic();
  topic.toLowerCase();
  topic.trim();
  String payload = pub.payload_string();
  payload.toLowerCase();
  payload.trim();

  if (topic.equals(mqttGetTopic(mqttSyncStatus))) {
    if (payload.equals(mqttSyncRequest)) {
      _callbackDeviceSync(topic, pub.payload_string());
    };
  }
  else if (topic.startsWith(mqttGetTopic(mqttSettings))) {
    _callbackSettings(topic, pub.payload_string());
  };

  _led->ledOff();
}

// =============================================================================================================================
// == Основной цикл ============================================================================================================
// =============================================================================================================================
void mqttLoop() 
{
  if (mqttConnect()) {
    _mqttClient.loop();

    if ((millis() - _lastPublishSensors >= _intvPublishSensors * 1000)) {
      mqttPublishSensors();
    };
  };

  _tsProcessQueue();
  
  _led->ledOff();
}
